// v8sharp is a community project available under the Microsoft Public License (Ms-PL)
// Code is provided as is and with no warrenty  Use at your own risk
// View the project and the latest code at http://v8sharp.codeplex.com/

#include "v8type.h"
#include "v8engineimpl.h"
#include "v8scriptimpl.h"
#include "v8scriptexception.h"
#include "v8function.h"

namespace v8sharp {
	ref class NamespaceFilter sealed {
	public:
		NamespaceFilter(String^ nameSpace) {
			this->nameSpace = nameSpace;
		}

		bool Include(Type^ type) {
			return type->Namespace->StartsWith(this->nameSpace);
		}

	private:
		String^ nameSpace;
	};

	V8EngineImpl::V8EngineImpl() {
		this->handleScope = new V8HandleScope();
		this->objectTemplate = new V8ObjectTemplate();

		this->registeredTypes = gcnew List<V8RegisteredType^>();
	}

	void V8EngineImpl::Register(String^ name, Object^ instance) {
		Register(name, instance, instance->GetType());
	}

	generic<typename T> 
	void V8EngineImpl::Register(String^ name, T instance) {
		Register(name, instance, T::typeid);
	}

	void V8EngineImpl::Register(String^ name, Object^ instance, Type^ type) {
		if (String::IsNullOrEmpty(name)) {
			throw gcnew ArgumentNullException("name");
		}
		if (instance == nullptr) {
			throw gcnew ArgumentNullException("instance");
		}
		if (type == nullptr) {
			throw gcnew ArgumentNullException("type");
		}
		if (!type->IsAssignableFrom(instance->GetType())) {
			throw gcnew ArgumentException();
		}

		V8RegisteredType^ registeredType = EnsureRegistered(type, true);
		registeredType->Register(name, instance);
	}

	void V8EngineImpl::RegisterFunction(String^ name, Delegate^ function) {
		v8::HandleScope handleScope;
		v8::Handle<v8::ObjectTemplate> tmpl;
		v8::Handle<v8::ObjectTemplate> protoTmpl;

		v8::Context::Scope contextScope(this->objectTemplate->GetContext());

		pin_ptr<const wchar_t> namePtr = PtrToStringChars(name);
		v8::Handle<v8::String> v8FunctionName = v8::String::New((const uint16_t*)namePtr);
		v8::Handle<v8::Function> v8Function = V8FunctionHandler::CreateFunction(function);
		this->objectTemplate->GetContext()->Global()->Set(v8FunctionName, v8Function, v8::ReadOnly);
	}

	generic<typename T> 
	void V8EngineImpl::Register() {
		Register(T::typeid);
	}

	void V8EngineImpl::Register(Type^ type) {
		EnsureRegistered(type);
	}

	void V8EngineImpl::Register(Assembly^ assembly) {
		this->Register(assembly, (Predicate<Type^>^)nullptr);
	}

	void V8EngineImpl::Register(Assembly^ assembly, Predicate<Type^>^ typeFilter) {
		for each(Type^ type in assembly->GetExportedTypes()) {
			if (type->IsAbstract || type->IsInterface) {
				continue;
			}
			if (typeFilter != nullptr && !typeFilter(type)) {
				continue;
			}

			Register(type);
		}
	}

	void V8EngineImpl::Register(Assembly^ assembly, String^ nameSpace) {
		NamespaceFilter^ filter = gcnew NamespaceFilter(nameSpace);
		Predicate<Type^>^ typeFilter = gcnew Predicate<Type^>(filter, &NamespaceFilter::Include);
		this->Register(assembly, typeFilter);
	}

	V8RegisteredType^ V8EngineImpl::EnsureRegistered(Type^ type, bool allowAbstractOrInterface) {
		if (type == nullptr) {
			throw gcnew ArgumentNullException("type");
		}
		if (!allowAbstractOrInterface && (type->IsAbstract || type->IsInterface)) {
			throw gcnew ArgumentException();
		}

		V8RegisteredType^ registeredType = nullptr;
		for (int i = 0, count = this->registeredTypes->Count; i < count; i++) {
			registeredType = this->registeredTypes[i];
			if (registeredType->Type != type) {
				registeredType = nullptr;
			} else {
				break;
			}
		}

		if (registeredType == nullptr) {
			registeredType = gcnew V8RegisteredType(type, this->objectTemplate);
			registeredTypes->Add(registeredType);
		}

		return registeredType;
	}

	Object^ V8EngineImpl::Execute(String^ script) {
		V8Script^ compiledScript = this->Compile(script);
		return this->Execute(compiledScript);
	}

	Object^ V8EngineImpl::Execute(V8Script^ script) {
		if (script == nullptr) {
			throw gcnew ArgumentNullException("script");
		}
		if (script->HasExceptions) {
			throw gcnew V8ScriptException(script->Exceptions);
		}

		return ((V8ScriptImpl^)script)->Execute(this->objectTemplate->GetContext());
	}

	V8Script^ V8EngineImpl::Compile(String^ script) {
		if (String::IsNullOrEmpty(script)) {
			throw gcnew ArgumentNullException("script");
		}

		return V8ScriptImpl::Compile(this->objectTemplate->GetContext(), script);
	}

	void V8EngineImpl::Destroy() {
		if (this->objectTemplate) {
			delete this->objectTemplate;
			this->objectTemplate = NULL;
		}
		if (this->handleScope) {
			delete this->handleScope;
			this->handleScope = NULL;
		}
	}

	V8Engine^ V8Engine::Create() {
		return gcnew V8EngineImpl();
	}
}