// v8sharp is a community project available under the Microsoft Public License (Ms-PL)
// Code is provided as is and with no warrenty  Use at your own risk
// View the project and the latest code at http://v8sharp.codeplex.com/

#include "v8external.h"
#include "v8type.h"
#include "v8value.h"

using namespace System::Reflection;

v8::Handle<v8::Value> V8ObjectConstructor(const v8::Arguments& info);

v8::Handle<v8::Value> V8MethodInvoker(const v8::Arguments& info);

v8::Handle<v8::Value> V8PropertyGetter(v8::Local<v8::String> propertyName, const v8::AccessorInfo& info);

void V8PropertySetter(v8::Local<v8::String> propertyName, v8::Local<v8::Value> value, const v8::AccessorInfo& info);

namespace v8sharp {	
	V8RegisteredType::V8RegisteredType(System::Type^ type, V8ObjectTemplate* rootTemplate) {
		this->instances = new V8InstanceList();
		this->methodList = new V8MethodList();

		this->type = type;
		this->objectTemplate = CreateObjectTemplate(this, rootTemplate);
		this->rootTemplate = rootTemplate;
	}

	System::Type^ V8RegisteredType::Type::get() {
		return this->type;
	}

	V8ObjectTemplate* V8RegisteredType::GetInternal() {
		return this->objectTemplate;
	}

	void V8RegisteredType::Destroy() {
		if (this->objectTemplate) {
			delete this->objectTemplate;
			this->objectTemplate = NULL;
		}
		if (this->instances) {
			delete this->instances;
			this->instances = NULL;
		}
		if (this->methodList) {
			delete this->methodList;
			this->methodList = NULL;
		}
	}

	v8::Handle<v8::Object> V8RegisteredType::GetObject(String^ nameSpace, V8ObjectTemplate* rootTemplate) {
		v8::Handle<v8::Object> obj = rootTemplate->GetContext()->Global();
		
		if (!String::IsNullOrEmpty(nameSpace)) {
			cli::array<String^>^ parts = nameSpace->Split(gcnew cli::array<Char>('.'));
			for each (String^ part in parts) {
				pin_ptr<const wchar_t> partPtr = PtrToStringChars(part);
				v8::Handle<v8::String> stringPtr = v8::String::New((const uint16_t*)partPtr);
				v8::Handle<v8::Value> partObj = obj->Get(stringPtr);

				if (partObj.IsEmpty() || partObj->IsUndefined()) {
					partObj = v8::Object::New();
					obj->Set(stringPtr, partObj, v8::ReadOnly);
					obj = v8::Handle<v8::Object>::Cast(partObj);
				} else if (partObj->IsObject()) {
					obj = v8::Handle<v8::Object>::Cast(partObj);
				} else {
					throw gcnew InvalidOperationException();
				}
			}
		}

		return obj;
	}

	V8ObjectTemplate* V8RegisteredType::CreateObjectTemplate(V8RegisteredType^ registeredType, V8ObjectTemplate* rootTemplate) {
		System::Type^ type = registeredType->Type;

		v8::HandleScope handleScope;
		v8::Handle<v8::ObjectTemplate> tmpl;
		v8::Handle<v8::ObjectTemplate> protoTmpl;

		{
			v8::Context::Scope contextScope(rootTemplate->GetContext());
			v8::Handle<v8::Object> obj = GetObject(type->Namespace, rootTemplate);
			
			pin_ptr<const wchar_t> namePtr = PtrToStringChars(type->Name);
			v8::Handle<v8::String> classPtr = v8::String::New((const uint16_t*)namePtr);
			
			v8::Handle<v8::FunctionTemplate> constructor;
			v8::Handle<v8::Value> external = v8::External::New((void*)V8ExternalWrapper::Create(registeredType));
			if (!(type->IsAbstract || type->IsInterface)) {				
				constructor = v8::FunctionTemplate::New(V8ObjectConstructor, external);
				constructor->SetClassName(classPtr);
		
				tmpl = constructor->InstanceTemplate();
				protoTmpl = constructor->PrototypeTemplate();
			} else {
				tmpl = v8::ObjectTemplate::New();
				protoTmpl = tmpl;
			}

			tmpl->SetInternalFieldCount(1);

			for each (PropertyInfo^ prop in type->GetProperties()) {
				v8::Handle<v8::String> v8PropertyName = v8::Handle<v8::String>::Cast(V8ValueWrapper::Unwrap(prop->Name));
				protoTmpl->SetAccessor(v8PropertyName, V8PropertyGetter, V8PropertySetter, external);
			}

			V8MethodList* methodList = registeredType->methodList;
			List<String^>^ uniqueMethodNames = gcnew List<String^>();
			for each (MethodInfo^ method in type->GetMethods()) {
				if (uniqueMethodNames->Contains(method->Name)) {
					continue;
				} else {
					V8MethodInvocation* methodInvocation = new V8MethodInvocation();
					methodInvocation->Name = method->Name;
					methodInvocation->RegisteredType = registeredType;
					methodList->Instance = methodInvocation;
					methodList->Next = new V8MethodList();
					methodList = methodList->Next;

					v8::Handle<v8::FunctionTemplate> v8Method = v8::FunctionTemplate::New(V8MethodInvoker, v8::External::New((void*)methodInvocation));
					v8::Handle<v8::String> v8MethodName = v8::Handle<v8::String>::Cast(V8ValueWrapper::Unwrap(method->Name));

					protoTmpl->Set(v8MethodName, v8Method->GetFunction());
				}
			}

			if (!(type->IsAbstract || type->IsInterface)) {
				obj->Set(classPtr, constructor->GetFunction());
			}
		}
				
		return new V8ObjectTemplate(tmpl);
	}

	void V8RegisteredType::Register(String^ name, Object^ instance) {
		v8::HandleScope handleScope;
		v8::Context::Scope contextScope(this->rootTemplate->GetContext());

		String^ nameSpace = nullptr;
		String^ instanceName = name;
		if (name->Contains(".")) {
			int index = name->LastIndexOf(".");
			nameSpace = name->Substring(0, index);
			instanceName = name->Substring(index + 1);
		}

		V8InstanceList* current = this->instances;
		while (current->IsFilled) {
			if (String::Equals(current->Name, name, StringComparison::OrdinalIgnoreCase)) {
				delete current->NativeInstance;
				current->IsFilled = false;
				break;
			}
			if (current->Next) {
				current = current->Next;
			} else {
				break;
			}
		}

		if (current->IsFilled) {
			current->Next = new V8InstanceList();
			current = current->Next;
		}

		current->IsFilled = true;
		current->Name = name;
		current->Instance = instance;
		current->NativeInstance = V8ExternalWrapper::Create(instance);

		v8::Handle<v8::Object> ns = GetObject(nameSpace, this->rootTemplate);
		v8::Handle<v8::External> ext = v8::External::New((void*)current->NativeInstance);

		v8::Handle<v8::Object> obj = this->GetInternal()->GetTemplate()->NewInstance();
		//if the type is creatable then an instance was already created from V8ObjectConstructor
		//clean it up
		if (!(type->IsAbstract || type->IsInterface)) {
			v8::Handle<v8::Value> tmpV8 = obj->GetInternalField(0);
			if (!tmpV8.IsEmpty() && tmpV8->IsExternal()) {
				V8ExternalWrapper* tmp = (V8ExternalWrapper*)v8::Handle<v8::External>::Cast(tmpV8)->Value();
				if (tmp) {
					delete tmp;
				}
			}
		}

		obj->SetInternalField(0, ext);
	
		pin_ptr<const wchar_t> namePtr = PtrToStringChars(instanceName);
		v8::Handle<v8::String> v8Name = v8::String::New((const uint16_t*)namePtr);
		ns->Set(v8Name, obj);
	}
}

v8::Handle<v8::Value> V8ObjectConstructor(const v8::Arguments& info) {
	v8::HandleScope handleScope;

	v8sharp::V8RegisteredType^ registeredType = safe_cast<v8sharp::V8RegisteredType^>(v8sharp::V8ValueWrapper::Wrap(info.Data()));
	Type^ type = registeredType->Type;

	int len = info.Length();
	cli::array<Object^>^ args = gcnew cli::array<Object^>(len);
	for (int i = 0; i < len; i++) {
		args[i] = v8sharp::V8ValueWrapper::Wrap(info[i]);
	}

	Object^ instance = System::Activator::CreateInstance(type, args);
	V8ExternalWrapper* ext = V8ExternalWrapper::Create(instance);
	
	info.This()->SetInternalField(0, v8::External::New((void*)ext));

	return v8::Handle<v8::Value>();
}

v8::Handle<v8::Value> V8MethodInvoker(const v8::Arguments& info) {
	v8::HandleScope handleScope;
	v8::Handle<v8::Value> returnValue = v8::Undefined();

	V8MethodInvocation* methodInvocation = (V8MethodInvocation*)v8::Handle<v8::External>::Cast(info.Data())->Value();
	v8sharp::V8RegisteredType^ registeredType = methodInvocation->RegisteredType;
	Type^ type = registeredType->Type;

	int len = info.Length();
	cli::array<Object^>^ args = gcnew cli::array<Object^>(len);
	cli::array<Type^>^ argTypes = gcnew cli::array<Type^>(len);
	for (int i = 0; i < len; i++) {
		Object^ arg = v8sharp::V8ValueWrapper::Wrap(info[i]);
		args[i] = arg;
		argTypes[i] = arg->GetType();
	}

	String^ methodName = methodInvocation->Name;
	MethodInfo^ method = type->GetMethod(methodName, argTypes);
	if (method != nullptr) {
		Object^ instance = v8sharp::V8ValueWrapper::Wrap(info.This()->GetInternalField(0));
		Object^ value = method->Invoke(instance, args);
		
		returnValue = handleScope.Close(v8sharp::V8ValueWrapper::Unwrap(value));
	}
	
	return returnValue;
}

v8::Handle<v8::Value> V8PropertyGetter(v8::Local<v8::String> propertyName, const v8::AccessorInfo& info) {
	v8::HandleScope handleScope;
	v8::Handle<v8::Value> returnValue = v8::Undefined();

	v8sharp::V8RegisteredType^ registeredType = safe_cast<v8sharp::V8RegisteredType^>(v8sharp::V8ValueWrapper::Wrap(info.Data()));
	Type^ type = registeredType->Type;

	String^ propName = safe_cast<String^>(v8sharp::V8ValueWrapper::Wrap(propertyName));
	PropertyInfo^ prop = type->GetProperty(propName);
	if (prop != nullptr && prop->CanRead) {
		Object^ instance = v8sharp::V8ValueWrapper::Wrap(info.This()->GetInternalField(0));
		if (instance != nullptr) {
			Object^ value = prop->GetValue(instance, nullptr);
		
			returnValue = handleScope.Close(v8sharp::V8ValueWrapper::Unwrap(value));
		}
	}
	
	return returnValue;
}

void V8PropertySetter(v8::Local<v8::String> propertyName, v8::Local<v8::Value> value, const v8::AccessorInfo& info) {
	v8::HandleScope handleScope;

	v8sharp::V8RegisteredType^ registeredType = safe_cast<v8sharp::V8RegisteredType^>(v8sharp::V8ValueWrapper::Wrap(info.Data()));
	Type^ type = registeredType->Type;

	String^ propName = safe_cast<String^>(v8sharp::V8ValueWrapper::Wrap(propertyName));
	PropertyInfo^ prop = type->GetProperty(propName);
	if (prop != nullptr && prop->CanWrite) {
		Object^ instance = v8sharp::V8ValueWrapper::Wrap(info.This()->GetInternalField(0));
		if (instance != nullptr) {		
			Object^ val = v8sharp::V8ValueWrapper::Wrap(value);
			prop->SetValue(instance, val, nullptr);
		}
	}
}

V8ObjectTemplate::V8ObjectTemplate(v8::Handle<v8::ObjectTemplate> objTemplate) {
	v8::HandleScope handleScope;

	if (objTemplate.IsEmpty()) {
		v8::Handle<v8::ObjectTemplate> objTemp = v8::ObjectTemplate::New();
		this->objectTemplate = v8::Persistent<v8::ObjectTemplate>::New(handleScope.Close(objTemp));

		v8::Handle<v8::Context> ctx = v8::Context::New(NULL, this->objectTemplate);
		this->context = v8::Persistent<v8::Context>::New(ctx);
	} else {
		this->objectTemplate = v8::Persistent<v8::ObjectTemplate>::New(objTemplate);

		v8::Handle<v8::Context> ctx = v8::Context::New();
		this->context = v8::Persistent<v8::Context>::New(ctx);
	}
}

V8ObjectTemplate::~V8ObjectTemplate() {
	this->context.Dispose();
	this->objectTemplate.Clear();
}

v8::Handle<v8::ObjectTemplate> V8ObjectTemplate::GetTemplate() {
	return this->objectTemplate;
}

v8::Persistent<v8::Context> V8ObjectTemplate::GetContext() {
	return this->context;
}

V8InstanceList::V8InstanceList() {
	this->IsFilled = false;
	this->NativeInstance = NULL;
	this->Next = NULL;
}

V8InstanceList::~V8InstanceList() {
	if (this->NativeInstance) {
		delete this->NativeInstance;
		this->NativeInstance = NULL;
	}

	if (this->Next) {
		delete this->Next;
		this->Next = NULL;
	}
}

V8MethodList::~V8MethodList() {
	//NOTE: not sure why but this was throwing memory corruption error
	//uncomment after issue is resolved

	//if (this->Instance) {
	//	delete this->Instance;
	//	this->Instance = NULL;
	//}

	//if (this->Next) {
	//	delete this->Next;
	//	this->Next = NULL;
	//}
}