// v8sharp is a community project available under the Microsoft Public License (Ms-PL)
// Code is provided as is and with no warrenty  Use at your own risk
// View the project and the latest code at http://v8sharp.codeplex.com/

#include "v8value.h"
#include "v8external.h"
#include "v8function.h"

using namespace System::Collections::Generic;

class V8External;

namespace v8sharp {
	Object^ V8ValueWrapper::Wrap(v8::Handle<v8::Value> value) {
		if (value->IsString()) {
			return WrapString(value);
		} else if (value->IsBoolean()) {
			return WrapBoolean(value);
		} else if (value->IsInt32()) {
			return WrapInt32(value);
		} else if (value->IsArray()) {
			return WrapArray(value);
		} else if (value->IsDate()) {
			return WrapDate(value);
		} else if (value->IsNumber()) {
			return WrapDouble(value);
		} else if (value->IsFunction()) {
			return WrapFunction(value);
		} else if (value->IsExternal()) {
			return WrapExternal(value);
		} else if (value->IsObject()) {
			return WrapObject(value);
		} else if (value->IsUndefined() || value->IsNull()) {
			return nullptr;
		}

		return nullptr;
	}

	v8::Handle<v8::Value> V8ValueWrapper::Unwrap(Object^ value) {
		if (value != nullptr) {
			Type^ type = value->GetType();
			if (type == String::typeid) {
				return UnwrapString(safe_cast<String^>(value));
			} else if (type == bool::typeid) {
				return UnwrapBoolean(safe_cast<bool>(value));
			} else if (type == int::typeid) {
				return UnwrapInt32(safe_cast<int>(value));
			} else if (type == double::typeid) {
				return UnwrapDouble(safe_cast<double>(value));
			} else if (type->IsArray) {
				return UnwrapArray(safe_cast<Array^>(value));
			} else if (type == DateTime::typeid) {
				return UnwrapDate(safe_cast<DateTime>(value));
			} else if (Delegate::typeid->IsAssignableFrom(type)) {
				return UnwrapFunction(safe_cast<Delegate^>(value));
			} else {
				return UnwrapExternal(value);
			}
		}

		return v8::Null();
	}

	String^ V8ValueWrapper::WrapString(v8::Handle<v8::Value> value) {
		v8::String::AsciiValue ascii(value);
		return gcnew String((const char*)*ascii);
	}

	v8::Handle<v8::Value> V8ValueWrapper::UnwrapString(String^ value) {
		pin_ptr<const wchar_t> ptr = PtrToStringChars(value);
		return v8::String::New((const uint16_t*)ptr);
	}

	bool V8ValueWrapper::WrapBoolean(v8::Handle<v8::Value> value) {
		bool val = value->BooleanValue();
		return val;
	}

	v8::Handle<v8::Value> V8ValueWrapper::UnwrapBoolean(bool value) {
		if (value) {
			return v8::True();
		} else {
			return v8::False();
		}
	}

	int V8ValueWrapper::WrapInt32(v8::Handle<v8::Value> value) {
		int val = value->Int32Value();
		return val;
	}

	v8::Handle<v8::Value> V8ValueWrapper::UnwrapInt32(int value) {
		return v8::Int32::New(value);
	}

	double V8ValueWrapper::WrapDouble(v8::Handle<v8::Value> value) {
		double val = value->NumberValue();
		return val;
	}

	v8::Handle<v8::Value> V8ValueWrapper::UnwrapDouble(double value) {
		return v8::Number::New(value);
	}

	cli::array<Object^>^ V8ValueWrapper::WrapArray(v8::Handle<v8::Value> value) {
		v8::Array* val = v8::Array::Cast(*value);
		unsigned int len = val->Length();

		cli::array<Object^>^ result = gcnew cli::array<Object^>(len);
		for (unsigned int i = 0; i < len; i++) {
			v8::Handle<v8::Value> key = v8::Uint32::New(i);
			result[i] = Wrap(val->Get(key));
		}

		return result;
	}

	v8::Handle<v8::Value> V8ValueWrapper::UnwrapArray(Array^ value) {
		int len = value->Length;
		v8::Handle<v8::Array> result = v8::Array::New();
		
		for (int i = 0; i < len; i++) {
			v8::Handle<v8::Value> key = v8::Int32::New(i);
			result->Set(key, Unwrap(value->GetValue(i)));
		}

		return result;
	}

	DateTime V8ValueWrapper::WrapDate(v8::Handle<v8::Value> value) {
		DateTime startDate = DateTime(1970, 1, 1);

		double milliseconds = value->NumberValue();
		TimeSpan ts = TimeSpan::FromMilliseconds(milliseconds);

		return DateTime(ts.Ticks + startDate.Ticks).ToLocalTime();
	}

	v8::Handle<v8::Value> V8ValueWrapper::UnwrapDate(DateTime value) {
		DateTime startDate = DateTime(1970, 1, 1);

		TimeSpan ts = TimeSpan::FromTicks(value.Ticks - startDate.Ticks);
		return v8::Date::New(ts.TotalMilliseconds);
	}

	Delegate^ V8ValueWrapper::WrapFunction(v8::Handle<v8::Value> value) {
		v8::Function* function = v8::Function::Cast(*value);
		return V8FunctionHandler::CreateDelegate(function);
	}

	v8::Handle<v8::Value> V8ValueWrapper::UnwrapFunction(Delegate^ value) {
		return V8FunctionHandler::CreateFunction(value);
	}

	Object^ V8ValueWrapper::WrapObject(v8::Handle<v8::Value> value) {
		v8::Handle<v8::Object> obj = value->ToObject();
		if (obj->InternalFieldCount() > 0) {
			v8::Handle<v8::Value> internalField = obj->GetInternalField(0);
			if (internalField->IsExternal()) {
				return WrapExternal(internalField);
			}
		}

		v8::Local<v8::Array> names = obj->GetPropertyNames();
		
		unsigned int len = names->Length();
		Dictionary<String^, Object^>^ results = gcnew Dictionary<String^, Object^>(len);
		for (unsigned int i = 0; i < len; i++) {
			v8::Handle<v8::Value> nameKey = v8::Uint32::New(i);
			v8::Handle<v8::Value> propName = names->Get(nameKey);
			v8::Handle<v8::Value> propValue = obj->Get(propName);

			String^ key = WrapString(propName);
			results[key] = Wrap(propValue);
		}

		return results;
	}

	v8::Handle<v8::Value> V8ValueWrapper::UnwrapObject(Object^ value) {
		v8::Handle<v8::Object> obj = v8::Object::New();

		if (value->GetType() == Dictionary<String^, Object^>::typeid) {
			Dictionary<String^, Object^>^ dict = safe_cast<Dictionary<String^, Object^>^>(value);
			for each(KeyValuePair<String^, Object^> pair in dict) {
				v8::Handle<v8::Value> key = Unwrap(pair.Key);
				v8::Handle<v8::Value> val = Unwrap(pair.Value);

				obj->Set(key, val);
			}
		} else {
			for each(System::Reflection::PropertyInfo^ prop in value->GetType()->GetProperties()) {
				v8::Handle<v8::Value> key = Unwrap(prop->Name);
				v8::Handle<v8::Value> val = Unwrap(prop->GetValue(value, nullptr));

				obj->Set(key, val);
			}
		}

		return obj;
	}

	Object^ V8ValueWrapper::WrapExternal(v8::Handle<v8::Value> value) {
		V8ExternalWrapper* wrapper = V8ExternalWrapper::Unwrap(value);
		return wrapper->Object;
	}

	v8::Handle<v8::Value> V8ValueWrapper::UnwrapExternal(Object^ value) {
		V8ExternalWrapper* wrapper = V8ExternalWrapper::Create(value);
		return v8::External::New((void*)wrapper);
	}
}